import os
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Dict, Any, Union, BinaryIO

from .integrity import get_file_integrity, get_reader_integrity
from .limited_reader import LimitedReader


class Type(Enum):
    """
    File Type
    """
    DIRECTORY = 1
    LINK = 2
    FILE = 3


@dataclass
class Metadata:
    """
    File metadata stored in the header.
    """
    # path in asar file
    path: Path
    type: Type = None
    # directory
    files: Dict[str, "Metadata"] = None
    # link
    link: Path = None
    # file
    unpacked: bool = False
    size: int = 0
    offset: int = 0
    integrity: Dict[str, Any] = None
    executable: bool = False
    # real file path or stream to read
    file_path: Path = None
    file_reader: Union[BinaryIO, LimitedReader] = None

    def set_link(self, link_in: Path):
        """
        Set the node as a link node
        :param link_in: the target path to which the symbolic link points
        """
        self.type = Type.LINK
        self.link = link_in

    def set_dir(self, unpacked: bool):
        """
        Set the node as a directory node
        :param unpacked: Exclude from packaging?
        """
        self.unpacked = unpacked
        self.type = Type.DIRECTORY
        self.files = {}

    def set_file(self, file_path: Path, unpacked: bool):
        """
        Set the node as a file node
        :param file_path: The actual file path, through which we read file size and data
        :param unpacked: Exclude from packaging?
        """
        self.unpacked = unpacked
        self.type = Type.FILE
        stat = file_path.stat()
        self.size = stat.st_size
        self.integrity = get_file_integrity(file_path)
        self.file_path = file_path
        self.executable = os.name != "nt" and (stat.st_mode & 0o100)

    def set_stream(self, file_reader: Union[BinaryIO, LimitedReader], unpacked: bool):
        """
        Set the node as a file stream node
        :param file_reader: The actual file stream, through which we read file size and data
        :param unpacked: Exclude from packaging?
        """
        self.unpacked = unpacked
        self.type = Type.FILE
        self.integrity, self.size = get_reader_integrity(file_reader)
        self.file_reader = file_reader

    def set_from_other(self, other: "Metadata"):
        """
        Set the node from other metadata, but not including files and offset
        :param other:
        """
        self.path = other.path
        self.type = other.type
        if self.type == Type.LINK:
            self.link = other.link
        elif self.type == Type.DIRECTORY:
            self.unpacked = other.unpacked
            self.files = {}
        else:
            self.unpacked = other.unpacked
            self.size = other.size
            self.integrity = other.integrity
            self.executable = other.executable
            self.file_path = other.file_path
            self.file_reader = other.file_reader

    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize to a formatted dict, usually the next step is to serialize to json
        """
        if self.type == Type.DIRECTORY:
            files = {}
            for key, value in self.files.items():
                files[key] = value.to_dict()
            return {"files": files}

        if self.type == Type.LINK:
            return {"link": str(self.link)}

        rst = {"size": self.size, "integrity": self.integrity}
        if self.unpacked:
            rst["unpacked"] = self.unpacked
        else:
            rst["offset"] = str(self.offset)
        if self.executable:
            rst["executable"] = self.executable
        return rst



import io
from typing import BinaryIO


class LimitedReader:
    """
    LimitedReader wraps a part of a reader into a reader.
    """
    _reader: BinaryIO
    _offset: int
    _size: int
    _remaining: int

    def __init__(self, reader: BinaryIO, offset: int, size: int):
        self._reader = reader
        self._offset = offset
        self._size = size
        self._remaining = size

    def read(self, n=-1):
        if n == -1 or n > self._remaining:
            n = self._remaining
        if n == 0:
            return b''
        chunk = self._reader.read(n)
        self._remaining -= len(chunk)
        return chunk

    def seek(self, offset, whence=io.SEEK_SET):
        end_offset = self._offset + self._size
        if whence == io.SEEK_SET:
            offset = self._offset + offset
        elif whence == io.SEEK_CUR:
            offset = self._reader.tell() + offset
        elif whence == io.SEEK_END:
            offset = end_offset + offset
        else:
            raise ValueError("invalid whence value")
        if not (self._offset <= offset < end_offset):
            raise ValueError("seek position is out of bounds")
        self._remaining = end_offset - offset
        return self._reader.seek(offset, whence)



import io
import shutil
from pathlib import Path

from asar import create_archive, extract_archive, AsarArchive


def cmp(f1: Path, f2: Path) -> bool:
    if f1.is_symlink() and f2.is_symlink():
        return f1.resolve().relative_to(src.resolve()) == f2.resolve().relative_to(dst.resolve())
    if f1.is_dir() and f2.is_dir():
        return f1.name == f2.name
    if f1.is_file() and f2.is_file():
        with f1.open("rb") as sf, f2.open("rb") as df:
            return sf.read() == df.read()
    return False


if __name__ == '__main__':
    src, asar, dst = Path("./testdata"), Path("./testdata.asar"), Path("./output")
    create_archive(src, asar, unpack="*.exe")
    shutil.rmtree(dst, ignore_errors=True)
    extract_archive(asar, dst)

    src_files, dst_files = list(src.rglob("*")), list(dst.rglob("*"))
    assert len(src_files) == len(dst_files)
    for s, d in zip(src_files, dst_files):
        assert s.relative_to(src) == d.relative_to(dst)
        assert cmp(s, d), f"file {s} and {d} is different"

    with AsarArchive(asar, mode="r") as archive:
        pathlist = archive.list()
        assert pathlist == sorted([f.relative_to(src) for f in src_files])
        for path in pathlist:
            src_file = src / path
            if src_file.is_dir():
                continue
            if src_file.is_symlink():
                src_file = src_file.resolve()
            assert src_file.read_bytes() == archive.read(path, follow_link=True), f"file {path} is different from src"

        new_asar = Path("./testdata.new.asar")
        added1_path, added2_path = Path("assets/added1.txt"), Path("assets/added2.txt")
        added2_text = "什么啊"

        with AsarArchive(new_asar, mode="w") as writer:
            writer.pack_other_asar(archive)
            writer.pack_file(path_in=added1_path, file_path=src / "f1.txt")
            writer.pack_stream(path_in=added2_path, file_reader=io.BytesIO(added2_text.encode("utf-8")))

    shutil.rmtree(dst, ignore_errors=True)
    extract_archive(new_asar, dst)

    assert cmp(src / "f1.txt", dst / added1_path)
    assert (dst / added2_path).read_text(encoding="utf-8") == added2_text

    src_files, dst_files = list(src.rglob("*")), list(dst.rglob("*"))
    dst_files.remove(dst / added1_path)
    dst_files.remove(dst / added2_path)
    assert len(src_files) == len(dst_files)
    for s, d in zip(src_files, dst_files):
        assert s.relative_to(src) == d.relative_to(dst)
        assert cmp(s, d), f"file {s} and {d} is different"



"""
integrity is an object consisting of a few keys:

- A hashing algorithm, currently only SHA256 is supported.
- A hex encoded hash value representing the hash of the entire file.
- An array of hex encoded hashes for the blocks of the file. i.e. for a blockSize of 4KB this array contains the hash of every block if you split the file into N 4KB blocks.
- A integer value blockSize representing the size in bytes of each block in the blocks hashes above
"""

import hashlib
from pathlib import Path
from typing import BinaryIO, Dict, Any, Tuple

ALGORITHM = "SHA256"
BLOCK_SIZE = 4 * 1024 * 1024


def get_file_integrity(file_path: Path) -> Dict[str, Any]:
    """
    Calc the integrity of a file.
    :param file_path: The file path.
    :return: integrity
    """
    with file_path.open("rb") as reader:
        return get_reader_integrity(reader)[0]


def get_reader_integrity(reader: BinaryIO) -> Tuple[Dict[str, Any], int]:
    """
    Calc the integrity of a file stream.
    :param reader: Input file stream, it should be readable and seekable.
    :return: (integrity, size)
    """
    hasher = hashlib.sha256()
    blocks = []
    size = 0
    reader.seek(0)
    while chunk := reader.read(BLOCK_SIZE):
        size += len(chunk)
        hasher.update(chunk)
        blocks.append(hashlib.sha256(chunk).hexdigest())
    return {
        "algorithm": ALGORITHM,
        "hash": hasher.hexdigest(),
        "blockSize": BLOCK_SIZE,
        "blocks": blocks,
    }, size



"""
Asar is a simple extensive archive format for Electron Archive, it works like tar that concatenates all files
together without compression, while having random access support.
"""

import fnmatch
import json
import shutil
import struct
from pathlib import Path
from typing import Dict, Any, BinaryIO, List

from .limited_reader import LimitedReader
from .metadata import Metadata, Type

UINT32_MAX = 2 ** 32 - 1


def align_int(i: int, alignment: int) -> int:
    """
    Align i to alignment, same as `i + (alignment - (i % alignment))`
    """
    return (i + alignment - 1) & ~(alignment - 1)


_modes = {"r": "rb", "w": "wb"}


class AsarArchive:
    """
    Class with methods to open, read, write, close asar files.
    """
    asar: Path
    asar_unpacked: Path
    _asar_io: BinaryIO
    _mode: str
    _header: Metadata
    _offset: int
    metas: List[Metadata]

    def __init__(self, asar: Path, mode: str):
        """
        Open the asar file with mode read 'r', write 'w'
        :param asar: the asar file path
        :param mode: read 'r' or write 'w'
        """
        self.asar = asar
        self.asar_unpacked = Path(f"{self.asar}.unpacked")
        self._mode = mode
        self._header = Metadata(path=Path(""), type=Type.DIRECTORY, files={})
        self.metas = []
        self._offset = 0
        if self._mode not in _modes:
            raise ValueError("AsarArchive requires mode 'r' or 'w'")
        # noinspection PyTypeChecker
        self._asar_io = self.asar.open(_modes[self._mode])

    def __enter__(self):
        if self._mode == "r":
            self._read_from_asar()
        return self

    def __exit__(self, *args):
        if self._mode == "w":
            self._write_to_asar()
        self._asar_io.close()

    def list(self):
        """Return a list of file paths in the archive."""
        return [meta.path for meta in self.metas]

    def pack_other_asar(self, other: "AsarArchive"):
        for meta in other.metas:
            node = self._search_node_from_path(meta.path)
            node.set_from_other(meta)
            if meta.type == Type.FILE and not meta.unpacked:
                node.offset = self._offset
                self._offset += node.size

    def pack(self, src: Path, unpack: str = None):
        """
        Add all files from src path to the asar archive.
        :param src: src path
        :param unpack: do not pack files matching glob <expression>
        """
        if not src:
            raise ValueError("src path not provided")
        if not src.is_absolute() or src.is_symlink():
            src = src.resolve()
        for path in src.iterdir():
            self._pack(path, src, unpack)

    def _pack(self, path: Path, src: Path, unpack: str):
        path_in = path.relative_to(src)
        if path.is_symlink():
            node = self._search_node_from_path(path_in)
            node.set_link(path.resolve().relative_to(src))
            if node.link.parts[0] == "..":
                raise ValueError(f'${path_in}: file "{node.link}" links out of the package')
        elif path.is_dir():
            node = self._search_node_from_path(path_in)
            node.set_dir(unpacked=unpack and fnmatch.fnmatch(path.name, unpack))
            for child in path.iterdir():
                self._pack(child, src, unpack)
        else:
            self.pack_file(path_in, path, should_unpack=unpack and fnmatch.fnmatch(path.name, unpack))

    def pack_stream(self, path_in: Path, file_reader: BinaryIO, should_unpack: bool = False):
        """
        Add one file stream to the asar archive.
        :param path_in: the path in asar archive.
        :param file_reader: the file stream.
        :param should_unpack: unpacked in asar? If not packed, it will be written to unpacked path
        :return:
        """
        dir_node = self._search_node_from_path(path_in.parent)
        node = self._search_node_from_path(path_in)
        node.set_stream(file_reader, unpacked=dir_node.unpacked or should_unpack)
        if not node.unpacked:
            if node.size > UINT32_MAX:
                raise ValueError(f"{path_in}: file size can not be larger than 4.2GB")
            node.offset = self._offset
            self._offset += node.size

    def pack_file(self, path_in: Path, file_path: Path, should_unpack: bool = False):
        """
        Add one file to the asar archive.
        :param path_in: the path in asar archive.
        :param file_path: the actual file path.
        :param should_unpack: unpacked in asar? If not packed, it will be written to unpacked path
        """
        dir_node = self._search_node_from_path(path_in.parent)
        node = self._search_node_from_path(path_in)
        node.set_file(file_path, unpacked=dir_node.unpacked or should_unpack)
        if not node.unpacked:
            if node.size > UINT32_MAX:
                raise ValueError(f"{path_in}: file size can not be larger than 4.2GB")
            node.offset = self._offset
            self._offset += node.size

    def read(self, path_in: Path, follow_link: bool = True):
        """
        Read one file from asar by random access
        :param path_in: path in asar
        :param follow_link: read the actual file if the path is a link
        :return: file bytes
        """
        node = self._search_node_from_path(path_in, create=False)
        if node.type == Type.DIRECTORY:
            return b""
        elif node.type == Type.LINK:
            return self.read(node.link) if follow_link else b""
        elif node.unpacked:
            if not node.file_path.exists():
                raise FileNotFoundError(node.file_path)
            with node.file_path.open("rb") as reader:
                return reader.read()
        else:
            node.file_reader.seek(0)
            return node.file_reader.read()

    def extract(self, dst: Path = None):
        """
        Extract the asar archive to the dst path.
        :param dst: dst dir path
        """
        dst = dst if dst else Path.cwd()
        dst.mkdir(parents=True, exist_ok=True)
        if not dst.is_absolute() or dst.is_symlink():
            dst = dst.resolve()
        for meta in self.metas:
            cur_dst = dst / meta.path
            if meta.type == Type.DIRECTORY:
                cur_dst.mkdir(parents=True, exist_ok=True)
            elif meta.type == Type.LINK:
                src_link = dst / meta.link
                try:
                    cur_dst.symlink_to(src_link)
                except FileExistsError:
                    cur_dst.unlink()
                    cur_dst.symlink_to(src_link)
            elif meta.unpacked:
                if not meta.file_path.exists():
                    raise FileNotFoundError(meta.file_path)
                shutil.copyfile(meta.file_path, cur_dst)
            else:
                with cur_dst.open(mode="wb") as writer:
                    meta.file_reader.seek(0)
                    shutil.copyfileobj(meta.file_reader, writer)

    def _read_from_asar(self):
        data_size, header_size, header_object_size, header_string_size = struct.unpack("<4I", self._asar_io.read(16))
        header = json.loads(self._asar_io.read(header_string_size).decode("utf-8"))
        self._offset = 8 + header_size
        self._parse_metadata(header, Path(""))

    def _parse_metadata(self, info: Dict[str, Any], path_in: Path):
        for name, child in info["files"].items():
            cur_path = path_in / name
            node = self._search_node_from_path(cur_path)
            if "files" in child:
                node.set_dir("unpacked" in child and bool(child["unpacked"]))
                self._parse_metadata(child, cur_path)
            elif "link" in child:
                node.set_link(Path(child["link"]))
            else:
                node.unpacked = "unpacked" in child and bool(child["unpacked"])
                node.type = Type.FILE
                node.integrity = child["integrity"]
                node.size = child["size"]
                if node.unpacked:
                    node.file_path = self.asar_unpacked / node.path
                else:
                    node.offset = int(child["offset"])
                    node.file_reader = LimitedReader(self._asar_io, self._offset + node.offset, node.size)

    def _write_to_asar(self):
        header_json = json.dumps(self._header.to_dict(), sort_keys=True, separators=(",", ":"), ensure_ascii=False)
        header_json = header_json.encode("utf-8")
        data_size = 4
        header_string_size = len(header_json)
        aligned_size = align_int(header_string_size, data_size)
        header_object_size = aligned_size + data_size
        header_size = header_object_size + data_size
        self._asar_io.write(struct.pack("<4I", data_size, header_size, header_object_size, header_string_size))
        self._asar_io.write(header_json)
        self._asar_io.write(b'\0' * (aligned_size - header_string_size))
        for metadata in self.metas:
            if metadata.type != Type.FILE:
                continue
            if metadata.unpacked:
                dst = self.asar_unpacked / metadata.path
                dst.parent.mkdir(parents=True, exist_ok=True)
                if metadata.file_path:
                    shutil.copyfile(metadata.file_path, dst)
                elif metadata.file_reader:
                    with dst.open("wb") as writer:
                        metadata.file_reader.seek(0)
                        shutil.copyfileobj(metadata.file_reader, writer)
            else:
                if metadata.file_path:
                    with open(metadata.file_path, "rb") as reader:
                        shutil.copyfileobj(reader, self._asar_io)
                elif metadata.file_reader:
                    metadata.file_reader.seek(0)
                    shutil.copyfileobj(metadata.file_reader, self._asar_io)

    def _search_node_from_path(self, path_in: Path, create: bool = True) -> Metadata:
        node = self._header
        name = path_in.name
        if not name:
            return node
        for part in path_in.parent.parts:
            if part != ".":
                node = node.files[part]
        if node.files.get(name) is None:
            if create:
                metadata = Metadata(path=path_in)
                node.files[name] = metadata
                self.metas.append(metadata)
            else:
                raise FileNotFoundError(path_in)
        return node.files[name]


def create_archive(dir_path: Path, asar_path: Path, unpack: str = None):
    """
    Pack the dir to create an asar archive.
    :param dir_path: The dir to pack.
    :param asar_path: The output asar path.
    :param unpack: do not pack files matching glob <expression>
    """
    with AsarArchive(asar_path, "w") as asar:
        asar.pack(dir_path, unpack)


def extract_archive(asar_path: Path, dir_path: Path):
    """
    Extract the asar archive to the dir.
    :param asar_path: The asar archive path to extract.
    :param dir_path: The output dir path.
    """
    with AsarArchive(asar_path, "r") as asar:
        asar.extract(dir_path)
